// Copyright (C) 2019  Nexedi SA and Contributors.
//                     Kirill Smelkov <kirr@nexedi.com>
//
// This program is free software: you can Use, Study, Modify and Redistribute
// it under the terms of the GNU General Public License version 3, or (at your
// option) any later version, as published by the Free Software Foundation.
//
// You can also Link and Combine this program with other software covered by
// the terms of any of the Free Software licenses or any of the Open Source
// Initiative approved licenses and Convey the resulting work. Corresponding
// source of such a combination shall include the source code for all other
// software used.
//
// This program is distributed WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// See COPYING file for full licensing terms.
// See https://www.nexedi.com/licensing for rationale and options.

package main

import (
	"context"

	"lab.nexedi.com/kirr/neo/go/zodb"
	"lab.nexedi.com/kirr/neo/go/zodb/btree"
)

// ΔFtail represents tail of revisional changes to files.
//
// It semantically consists of
//
//	[]δF			; rev ∈ (tail, head]
//
// where δF represents a change in files space
//
//	δF:
//		.rev↑
//		{} file -> {}blk
//
// Only files and blocks explicitly requested to be tracked are guaranteed to
// be present. In particular a block that was not explicitly requested to be
// tracked, even if it was changed in δZ, is not guaranteed to be present in δF.
//
// ΔFtail provides the following operations:
//
//   .Track(file, blkTreePath)		- add file and block reached via BTree path to tracked set.
//
//   .Update(δZ) -> δF				- update files δ tail given raw ZODB changes
//   .ForgetPast(revCut)			- forget changes past revCut
//   .SliceByRev(lo, hi) -> []δF		- query for all files changes with rev ∈ (lo, hi]
//   .SliceByFileRev(file, lo, hi) -> []δfile	- query for changes of file with rev ∈ (lo, hi]
//   .LastBlkRev(file, #blk, at)		- query for what is last revision that changed
//						  file[#blk] as of @at database state.
//
// XXX δfile:
//	.rev↑
//	[]blk
//
// XXX concurrent use
//
// XXX see also zodb.ΔTail
type ΔFtail struct {
	// ΔFtail merge btree.ΔTail with history of ZBlk
	δBtail  *ΔBtail
	fileIdx map[*btree.LOBTree]SetBigFile	// root -> {} BigFile	XXX root -> oid?
}

// ΔF represents a change in files space.
type ΔF struct {
	Rev    zodb.Tid
	Change map[*BigFile]SetI64 // file -> δfile (= {}#blk)
}

// Δfile represents a change to one file.
type Δfile struct {
	Rev    zodb.Tid
	Change SetI64 // changed blocks
}

// NewΔFtail creates new ΔFtail object.
//
// Initial tracked set is empty.
// Initial coverage of created ΔFtail is (at₀, at₀].
func NewΔFtail(at0 zodb.Tid) *ΔFtail {
	return &ΔFtail{
		δBtail:  NewΔBtail(at0),
		fileIdx: make(map[*btree.LOBTree]SetBigFile),
	}
}

// (tail, head] coverage
func (δFtail *ΔFtail) Head() zodb.Tid { return δFtail.δBtail.Head() }
func (δFtail *ΔFtail) Tail() zodb.Tid { return δFtail.δBtail.Tail() }


// Track adds tree path to tracked set and associates path root with file.
//
// XXX text
//
// A root can be associated with several files (each provided on different Track call).
func (δFtail *ΔFtail) Track(file *BigFile, path []btree.LONode) {
	δFtail.δBtail.Track(path)
	root := path[0].(*btree.LOBTree)
	files, ok := δFtail.fileIdx[root]
	if !ok {
		files = SetBigFile{}
		δFtail.fileIdx[root] = files
	}
	files.Add(file)

	// XXX mark something dirty so that LastBlkRev and Slice* know what to rebuild?

	// XXX debug
/*
	leaf := path[len(path)-1].(*btree.LOBucket)
	for _, e := range leaf.Entryv() { // XXX activate
		δFtail.tracked.Add(e.Key())
	}
*/
}

// Update updates δFtail given raw ZODB changes.
//
// It returns change in files space that corresponds to δZ.
//
// δZ should include all objects changed by ZODB transaction.
func (δFtail *ΔFtail) Update(δZ *zodb.EventCommit) ΔF {
	δB := δFtail.δBtail.Update(δZ)
	δF := ΔF{Rev: δB.Rev, Change: make(map[*BigFile]SetI64)}
	for root, δt := range δB.Change {
		files := δFtail.fileIdx[root]
		if len(files) == 0 {
			panicf("ΔFtail: root<%s> -> ø file", root.POid())
		}
		for file := range files {
			δfile, ok := δF.Change[file]
			if !ok {
				δfile = make(SetI64)
				δF.Change[file] = δfile
			}
			for blk /*, zblk*/ := range δt {
				// FIXME stub - need to take both keys and zblk changes into account
				// XXX document, and in particular how to include atTail
				δfile.Add(blk)
			}
		}
	}

	return δF
}

// ForgetPast discards all δFtail entries with rev ≤ revCut.
func (δFtail *ΔFtail) ForgetPast(revCut zodb.Tid) {
	panic("TODO")
}

// XXX
func (δFtail *ΔFtail) SliceByRev(lo, hi zodb.Tid) /*readonly*/ []ΔF {
	δassertSlice(δFtail, lo, hi)

	panic("TODO")
}

// SliceByFileRev returns history of file changes in (lo, hi] range.
//
// it must be called with the following condition:
//
//	tail ≤ lo ≤ hi ≤ head
//
// the caller must not modify returned slice.
//
// Note: contrary to regular go slicing, low is exclusive while high is inclusive.
func (δFtail *ΔFtail) SliceByFileRev(file *BigFile, lo, hi zodb.Tid) /*readonly*/[]Δfile {
	δassertSlice(δFtail, lo, hi)

	// merging tree (δT) and Zblk (δZblk) histories into file history (δFile):

	// δT		────────·──────────────·─────────────────·────────────
	//		        │              │
	//		        ↓              │
	// δZblk₁	────────────────o───────────────────o─────────────────
	//                                     |
	//                                     ↓
	// δZblk₂	────────────x────────────────x────────────────────────
	//
	//
	// δFile	────────o───────o──────x─────x────────────────────────

	vδZ := δFtail.δBtail.δZtail.SliceByRev(lo, hi)

	// FIXME stub: assumes blktab topology does not change.
	//       takes only ZBlk history into account.
	var vδfile []Δfile

/*
	// XXX activate zfile?
	vδT := δFtail.δBtail.SliceByRootRev(file.zfile.blktab, lo, hi)

	// state of `{} blk -> zblk` as we are scanning ↓
	δblktab := map[int64]struct {
		zblk   zodb.Oid // blk points to this zblk
		lo, hi zodb.Tid // blk points to zblk during [lo, hi)
	}{}

	iz := len(vδZ) - 1
	it := len(vδT) - 1
	for (iz >= 0 && it >= 0) {	// XXX -> ||
		δZ := vδZ[iz]
		δT := vδT[it]
		if δZ.Rev >= δT.Rev {
			for _, oid := range δZ.Changev {
				// XXX oid -> tracked ZBlk?
				// ZBlk -> bound to {}blk @head
				for blk := range boundToAtHead {
					if !δblktab.Has(blk) {
						δblktab[blk] = oid
					}
				}
			}
		}
		if δT.Rev >= δZ.Rev {
			...
		}
	}
*/

}

// LastBlkRev returns last revision that changed file[blk] as of @at database state.
//
// if exact=False - what is returned is only an upper bound for last block revision.
//
// f must be from head/
// at must ∈ (tail, head]	XXX [tail ?
// blk must be tracked
//
// XXX +ctx, error	rebuild []δF here
func (f *BigFile) LastBlkRev(ctx context.Context, blk int64, at zodb.Tid) (_ zodb.Tid, exact bool) {
	//defer xerr.Contextf(&err, "")	// XXX text

	δFtail := f.head.bfdir.δFtail

	// XXX tabRev -> treeRev ?
	// XXX activate zfile?
	zblkOid, ok, tabRev, tabRevExact, err := δFtail.δBtail.Get(ctx, f.zfile.blktab, blk, at)
	if err != nil {
		panic(err)
	}

	// block was removed
	// XXX or not in tracked set?
	if !ok {
		return tabRev, tabRevExact
	}

	// blktab[blk] was changed to point to a zblk @rev.
	// blk revision is max rev and when zblk changed last in (rev, at] range.
	//
	// XXX need to use full δZ, not only connected to tracked subset?
	zblkRev, zblkRevExact := δFtail.δBtail.δZtail.LastRevOf(zblkOid, at)
	if zblkRev > tabRev {
		return zblkRev, zblkRevExact
	} else {
		return tabRev, tabRevExact
	}
}
